#define _GNU_SOURCE

#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <math.h>
#include <getopt.h>
#include <stdbool.h>
#include <signal.h>
#include <unistd.h>

#include <assert.h>
#include <SDL/SDL.h>
#include "types.h"

#include "replay.h"

#define PLAY_HVL_VERSION "1.3"

#define FREQ 48000
#define HIVELY_LEN FREQ/50
#define OUTPUT_LEN 4096

int16 hivelyLeft[HIVELY_LEN], hivelyRight[HIVELY_LEN];
size_t hivelyIndex;
struct hvl_tune *tune;
bool autogain = false;
uint32 max_frames = 10 * 60 * 50;
bool use_songend = true;
volatile bool stop = false;
uint32 frames = 0;

static void sighandler( int s )
{
    if ( s == SIGINT ) {
        stop = true;
    }
}

void mix_and_play( struct hvl_tune *ht, uint8 *stream, int length );

BOOL init( void )
{
  SDL_AudioSpec wanted;

  if(SDL_Init(SDL_INIT_AUDIO)< 0) {
    printf("Could not initialize SDL: %s\n", SDL_GetError());
    SDL_Quit();
    return FALSE;
  }

  wanted.freq = FREQ; 
  wanted.format = AUDIO_S16SYS; 
  wanted.channels = 2; /* 1 = mono, 2 = stereo */
  wanted.samples = OUTPUT_LEN; // HIVELY_LEN;

  wanted.callback = (void*) mix_and_play;
  wanted.userdata = tune;

  if(SDL_OpenAudio(&wanted, NULL) < 0) {
    printf("Failed to open audio device.\n");
    SDL_Quit();
    return FALSE;
  }

  return TRUE;
}

void mix_and_play( struct hvl_tune *ht, uint8 *stream, int length )
{
    int16 *out;
    int i;
    size_t streamPos = 0;
    length = length >> 1;

    // Mix to 16bit interleaved stereo
    out = (int16*) stream;
    // Flush remains of previous frame
    for(i = hivelyIndex; i < (HIVELY_LEN); i++) {
        out[streamPos++] = hivelyLeft[i];
        out[streamPos++] = hivelyRight[i];
    }

    if ( ! tune ||
         stop ||
         ( use_songend && tune->ht_SongEndReached ) ||
         ( !use_songend && frames >= max_frames ) ) {
        memset( stream + streamPos, 0, length - streamPos);
        hivelyIndex = HIVELY_LEN;
        stop = true;
        return;
    }

    while(streamPos < length) {
        hvl_DecodeFrame( tune, (int8 *) hivelyLeft, (int8 *) hivelyRight, 2 );
        frames++;
        for(i = 0; i < (HIVELY_LEN) && streamPos < length; i++) {
            out[streamPos++] = hivelyLeft[i];
            out[streamPos++] = hivelyRight[i];
        }
    }
    hivelyIndex = i;
}

void print_help( int argc, char *argv[] )
{
    printf( "%s VERSION %s\n", argv[0], PLAY_HVL_VERSION );
    printf( "%s [options] <file name>\n", argv[0] );
    printf( "  -t m:ss\n  --time m:ss\n\t\tStop playing after given time\n" );
    printf( "  -a\n  --auto-gain\n\t\tNormalize output to best gain\n" );
}

int main( int argc, char *argv[] )
{
    int i;
    hivelyIndex = HIVELY_LEN;
    bool showhelp = false;

    while ( true ) {
        int option_index;
        static struct option long_options[] = {
                                               { "help", no_argument, NULL, 'h' },
                                               { "version", no_argument, NULL, 'V' },
                                               { "auto-gain", no_argument, NULL, 'a' },
                                               { "restore-tabs", optional_argument, NULL, 't' },
                                               { NULL, 0, NULL, 0 }
        };

        int c = getopt_long( argc, argv, "hVat:", long_options, &option_index );

        if ( c == -1 ) break;
      
        switch( c ) {
            case 'a':
                autogain = true;
                break;
            case 't':
                if ( optarg != NULL ) {
                    int m, s;

                    if ( sscanf( optarg, "%d:%d", &m, &s ) != 2 ) {
                        showhelp = true;
                    } else {
                        max_frames = m * 60 + s;
                        max_frames *= 50;
                        use_songend = false;
                    }
                } else {
                    showhelp = true;
                }
                break;
            case 'h':
            case 'V':
                showhelp = true;
                break;
            default:
                break;
        }
    }

    if ( optind >= argc || showhelp ) {
        print_help( argc, argv );
        return 0;
    }

    struct sigaction sig_ac;
  
    memset( &sig_ac, 0, sizeof( sig_ac ) );
    sig_ac.sa_handler = sighandler;
    sigemptyset( &sig_ac.sa_mask );

    sigaction( SIGINT, &sig_ac, NULL );

    if( init() ) {
        hvl_InitReplayer();
        tune = hvl_LoadTune( (TEXT*)argv[optind], FREQ, 4 );

        if( tune ) {
            hvl_InitSubsong( tune, 0 );

            if( autogain ) {
                printf( "Calculating optimal gain...\n" );
                int32 l = hvl_FindLoudest( tune, max_frames, use_songend );
                if ( l > 0 ) {
                    l = 3276700 / l;
                } else {
                    l = 100;
                }
    
                if ( l > 200 ) l = 200;
                tune->ht_mixgain = ( l * 256 ) / 100;

                printf( "Gain is %2.2f\n", (double)l / 100.0 );

                hvl_InitSubsong( tune, 0 );
                tune->ht_SongEndReached = 0;
            }

            printf( "Songname: %s\nChannels: %i\n", tune->ht_Name, tune->ht_Channels );
            printf( "Instruments:\n" );
            for( i=1; i <= tune->ht_InstrumentNr; i++ ) {
                printf( "%02d: %s\n", i, tune->ht_Instruments[i].ins_Name );
            }

            SDL_PauseAudio(0);

            while ( ! stop ) {
                usleep( 10 * 1000 );
            }

            SDL_PauseAudio(1);

            hvl_FreeTune( tune );
        }
    }
    SDL_Quit();
    return 0;
}
